@@ -1,9 +1,14 @@
+import os
 import torch
 import numpy as np
 from torch.utils.data import DataLoader, random_split
-# import trainer and evaluator engine
+# import trainer, evaluator and metrics
 from active.helpers.trainer import Trainer
 from active.helpers.evaluator import Evaluator
+from active.helpers.metrics import (
+    SeqEvalMetrics, 
+    NestedSeqEvalMetrics
+)
 # import data processors
 from active.helpers.processor import (
     BioTaggingProcessor,
@@ -11,32 +16,17 @@ from active.helpers.processor import (
     SequenceClassificationProcessor
 )
 # import ignite
-from ignite.engine import Events
+from ignite.engine import Engine, Events
 from ignite.contrib.handlers.tqdm_logger import ProgressBar
 from ignite.metrics import Recall, Precision, Average, Fbeta, Accuracy
 # hugginface
 import datasets
 import transformers
 # others
-import wandb
 from enum import Enum
 from typing import Tuple, Dict, Union, Callable, Type
 from pydantic import BaseModel, validator, root_validator
 
-def attach_metrics(engine, tag=None):
-    # create metrics
-    L = Average(output_transform=lambda out: out['loss'])
-    A = Accuracy(output_transform=type(engine).get_logits_and_labels)
-    R = Recall(output_transform=type(engine).get_logits_and_labels, average=True)
-    P = Precision(output_transform=type(engine).get_logits_and_labels, average=True)
-    F = Fbeta(beta=1.0, output_transform=type(engine).get_logits_and_labels, average=True)
-    # attach metrics
-    L.attach(engine, 'L' if tag is None else ('%s/L' % tag))
-    A.attach(engine, 'A' if tag is None else ('%s/A' % tag))
-    R.attach(engine, 'R' if tag is None else ('%s/R' % tag))
-    P.attach(engine, 'P' if tag is None else ('%s/P' % tag))
-    F.attach(engine, 'F' if tag is None else ('%s/F' % tag))
-
 class ForwardForNestedTokenClassification(transformers.PreTrainedModel):
     """ Forward container class used by transformer models for nested token classification.
         Has to be combined with a transformer model for token classification.
@@ -245,6 +235,45 @@ class DataConfig(BaseModel):
         
         # return preprocessed datasets
         return ds
+    
+    def attach_metrics(self, engine:Engine, tag:str) -> None:
+        # create token-level metrics
+        L = Average(output_transform=lambda out: out['loss'])
+        A = Accuracy(output_transform=type(engine).get_logits_and_labels)
+        R = Recall(output_transform=type(engine).get_logits_and_labels, average=True)
+        P = Precision(output_transform=type(engine).get_logits_and_labels, average=True)
+        F = Fbeta(beta=1.0, output_transform=type(engine).get_logits_and_labels, average=True)
+        
+        if self.task is Task.SEQUENCE:
+            # attach metrics
+            L.attach(engine, '%s/L' % tag)
+            A.attach(engine, '%s/A' % tag)
+            R.attach(engine, '%s/R' % tag)
+            P.attach(engine, '%s/P' % tag)
+            F.attach(engine, '%s/F' % tag)
+
+        elif self.task in (Task.BIO_TAGGING, Task.NESTED_BIO_TAGGING):
+            # attach token-level metrics
+            L.attach(engine, '%s/token/L' % tag)
+            A.attach(engine, '%s/token/A' % tag)
+            R.attach(engine, '%s/token/R' % tag)
+            P.attach(engine, '%s/token/P' % tag)
+            F.attach(engine, '%s/token/F' % tag)
+            
+            # create and attach sequence level metric
+            if self.task is Task.BIO_TAGGING:
+                SeqEvalMetrics(
+                    label_space=self.label_space, 
+                    output_transform=type(engine).get_logits_labels_mask
+                ).attach(engine, '%s/entity' % tag)
+
+            elif self.task is Task.NESTED_BIO_TAGGING:
+                NestedSeqEvalMetrics(
+                    # note that here the label space is the entity types
+                    entity_types=self.label_space,
+                    output_transform=type(engine).get_logits_labels_mask
+                ).attach(engine, '%s/entity' % tag)
+
 
 class ModelConfig(BaseModel):
     """ Model Configuration Model """
@@ -260,7 +289,7 @@ class ModelConfig(BaseModel):
         except OSError as e:
             # handle model invalid
             raise ValueError("Unkown pretrained checkpoint: %s" % value) from e
- 
+
         return value
 
     @property
@@ -335,26 +364,19 @@ class ExperimentConfig(BaseModel):
     def build_trainer(self) -> Trainer:
         return self.trainer.build_trainer(self.load_model())
 
-def main():
-    from argparse import ArgumentParser
-    # build argument parser
-    parser = ArgumentParser(description="Train transformer model on sequence or token classification tasks.")
-    parser.add_argument("--config", type=str, required=True, help="Path to a valid experiment configuration")
-    parser.add_argument("--use-cache", action='store_true', help="Load cached preprocessed datasets if available")
-    parser.add_argument("--seed", type=int, default=1337, help="Random Seed")
-    # parse arguments
-    args = parser.parse_args()
 
-    # set random seed
-    torch.manual_seed(args.seed)
-    np.random.seed(args.seed)
-
-    # parse configuration
-    config = ExperimentConfig.parse_file(args.config)
+def train(config:str, seed:int, use_cache:bool, disable_tqdm:bool =False):
+    import wandb
+   
+    config = ExperimentConfig.parse_file(config)
     print("Config:", config.json(indent=2))
 
+    # set random seed
+    torch.manual_seed(seed)
+    np.random.seed(seed)
+    
     # load datasets and build trainer
-    ds = config.load_dataset(use_cache=args.use_cache)
+    ds = config.load_dataset(use_cache=use_cache)
     trainer = config.build_trainer()
     
     # split validation dataset from train set
@@ -371,21 +393,26 @@ def main():
  
     # set up wandb 
     wandb_config = config.dict(exclude={"active"})
-    wandb_config['data']['dataset'] = config.data.dataset_info.builder_name
-    wandb_config['seed'] = args.seed
-    wandb.init(config=wandb_config)
+    wandb_config['data']['dataset'] = ds['train'].info.builder_name
+    wandb_config['seed'] = seed
+    wandb.init(
+        config=wandb_config,
+        project=os.environ.get("WANDB_PROJECT", "train-final"),
+        group=config.name
+    )
 
     # create the validater and tester
     validator = Evaluator(trainer.unwrapped_model)
     tester = Evaluator(trainer.unwrapped_model)
     # attach metrics
-    attach_metrics(trainer, tag="train")
-    attach_metrics(validator, tag="val")
-    attach_metrics(tester, tag="test")
+    config.data.attach_metrics(trainer, tag="train")
+    config.data.attach_metrics(validator, tag="val")
+    config.data.attach_metrics(tester, tag="test")
     # attach progress bar
-    ProgressBar(ascii=True).attach(trainer, output_transform=lambda output: {'L': output['loss']})
-    ProgressBar(ascii=True, desc='Validating').attach(validator)
-    ProgressBar(ascii=True, desc='Testing').attach(tester)
+    if not disable_tqdm:
+        ProgressBar(ascii=True).attach(trainer, output_transform=lambda output: {'L': output['loss']})
+        ProgressBar(ascii=True, desc='Validating').attach(validator)
+        ProgressBar(ascii=True, desc='Testing').attach(tester)
     
     # add event handler for testing and logging
     @trainer.on(Events.EPOCH_COMPLETED)
@@ -395,16 +422,35 @@ def main():
         test_state = tester.run(test_loader)
         # log both train and test metrics
         print("Step:", engine.state.iteration, "-" * 15)
-        print("Train F-Score:", engine.state.metrics['train/F'])
-        print("Val   F-Score:", val_state.metrics['val/F'])
-        print("Test  F-Score:", test_state.metrics['test/F'])
+        if config.task is Task.SEQUENCE:
+            print("Train F-Score:", engine.state.metrics['train/F'])
+            print("Val   F-Score:", val_state.metrics['val/F'])
+            print("Test  F-Score:", test_state.metrics['test/F'])
+        elif config.task in (Task.BIO_TAGGING, Task.NESTED_BIO_TAGGING):
+            print("Train Entity F-Score:", engine.state.metrics['train/entity/weighted avg/F'])
+            print("Val   Entity F-Score:", val_state.metrics['val/entity/weighted avg/F'])
+            print("Test  Entity F-Score:", test_state.metrics['test/entity/weighted avg/F'])
+        # log all metrics to weights and biases 
         wandb.log(
             test_state.metrics | val_state.metrics | engine.state.metrics, 
             step=engine.state.iteration
         )
 
     # run training
-    trainer.run(train_loader, **config.trainer.run_kwargs)
-   
+    return trainer.run(train_loader, **config.trainer.run_kwargs)
+
+def main():
+    from argparse import ArgumentParser
+    # build argument parser
+    parser = ArgumentParser(description="Train transformer model on sequence or token classification tasks.")
+    parser.add_argument("--config", type=str, required=True, help="Path to a valid experiment configuration")
+    parser.add_argument("--use-cache", action='store_true', help="Load cached preprocessed datasets if available")
+    parser.add_argument("--seed", type=int, default=1337, help="Random Seed")
+    # parse arguments
+    args = parser.parse_args()
+    
+    # train model
+    train(**vars(args))
+
 if __name__ == '__main__':
     main() 
